//*****************************************************************************
//*
//*
//*		PingApi.cpp
//*
//*
//*****************************************************************************
//
//	(C) Copyright Anton Zechner 2007
//
#include	<windows.h>
#include	<string.h>
#include	"PingApi.h"

#pragma		comment(lib, "wsock32.lib")

#if		   _MSC_VER >= 1000
#pragma 	once
#pragma		warning( disable :4514 )
#pragma 	warning( disable :4996 )
#endif

#define		STATUS_FAILED	0xFFFF
#define 	DEF_PACKET_SIZE 32
#define 	MAX_PACKET		1024

#define 	ICMP_ECHO		8
#define 	ICMP_ECHOREPLY	0
#define 	ICMP_MIN		8								// minimum 8 byte icmp packet (just header)


typedef struct 												// The IP header
	{
	unsigned char 		ucHdrLenVersion;    				// Version of IP / length of the header
	unsigned char		ucTypeOfService;					// Type of service
	unsigned short		usTotalLen;      					// total length of the packet
	unsigned short		usIdent;          					// unique identifier
	unsigned short		usFragFlags;						// flags
	unsigned char		ucTtl;
	unsigned char		ucProto;							// protocol (TCP, UDP etc)
	unsigned short		usChecksum;							// IP checksum
	unsigned int 		uSourceIp;
	unsigned int 		uDestIp;
	}IpHeader;

typedef struct
	{
	unsigned char 		bType;
	unsigned char 		bCode;								// type sub code
	unsigned short		usChkSum;
	unsigned short		usId;
	unsigned short		usSequ;
	unsigned int		uiTimeStamp;
	}IcmpHeader;


//*****************************************************************************
//*
//*		PingCmpIpAddr
//*
//*****************************************************************************
//	Vergleicht zwei IP-Adressen
//	pFrom		: Zeiger auf erste Adressse
//	pAddr		: Zeiger auf zweite Adressse
//	Ergibt 0 wenn sie gleich sind
static int PingCmpIpAddr(struct sockaddr_in *pFrom,struct sockaddr_in *pAddr)
{

  if(pFrom->sin_family     !=pAddr->sin_family    )return 1;
  if(pFrom->sin_addr.s_addr!=pAddr->sin_addr.s_addr)return 1;


return 0;
}


//*****************************************************************************
//*
//*		PingDecodeResponse
//*
//*****************************************************************************
// The response is an IP packet. We must decode the IP header to locate
// the ICMP data
static int PingDecodeResponse(char *pBuffer, int iBytes,struct sockaddr_in *pFrom,unsigned *pNumber)
{
IpHeader		*pIpHeader;
IcmpHeader		*pIcmpHeader;
unsigned short	 usHeaderLen;



	pIpHeader   = (IpHeader *)pBuffer;
	usHeaderLen = ( pIpHeader->ucHdrLenVersion&0x0F)*4;		// number of 32-bit words *4 = iBytes
	pIcmpHeader = (IcmpHeader*)(pBuffer + usHeaderLen);


	if(pIcmpHeader->bType != ICMP_ECHOREPLY)
		{
		*pNumber = 0xFFFFFFFF;
		return -3;											// non-echo type recvived
		}

	if(pIcmpHeader->usId != (USHORT)GetCurrentProcessId())
		{
		*pNumber = 0xFFFFFFFF;
		return -4;											// someone else's packet
		}

	*pNumber = pIcmpHeader->usSequ;



return pIcmpHeader->uiTimeStamp;
}

//*****************************************************************************
//*
//*		PingCheckSum
//*
//*****************************************************************************
static USHORT PingCheckSum(USHORT *pBuffer, int iSize)
{
unsigned long ulChkSum=0;


	while(iSize >1)
		{
		ulChkSum += *pBuffer++;
		iSize    -=  sizeof(USHORT);
		}

	if(iSize)
		{
		ulChkSum += *(UCHAR*)pBuffer;
		}

	ulChkSum  = (ulChkSum >> 16) + (ulChkSum & 0xFFFF);
	ulChkSum += (ulChkSum >> 16);


return (USHORT)~ulChkSum;
}

//*****************************************************************************
//*
//*		PingFilIcmpData
//*
//*****************************************************************************
// Helper function to fill in various stuff in our ICMP request.
static void PingFilIcmpData(char * pIcmpData, int iDataSize)
{
IcmpHeader	*pIcmpHdr;
char		*pDataPart;


	pIcmpHdr			= (IcmpHeader*)pIcmpData;
	pIcmpHdr->bType		= ICMP_ECHO;
	pIcmpHdr->bCode		= 0;
	pIcmpHdr->usId		= (USHORT)GetCurrentProcessId();
	pIcmpHdr->usChkSum	= 0;
	pIcmpHdr->usSequ	= 0;
	pDataPart			= pIcmpData + sizeof(IcmpHeader);

	memset(pDataPart,'E',iDataSize-sizeof(IcmpHeader));

}


//*****************************************************************************
//*
//*		PingInit
//*
//*****************************************************************************
//	Erstellt eine Socket-Verbindung und initialisiert die Daten
//	pData		: Zeiger auf die Ping Daten
//	pHostName	: Hostname oder IpAdresse die angepingt werden soll
//	iCount		: Ist die Anzahl der Ping-Paket-Eintrge
//	Ergibt 1 wenn ok sonst 0
int PingInit(PING_DATA *pData,const char *pHostName,int iCount)
{
WSADATA				sWsaData;
int					iTimeout;
static int			iInitDone=0;
struct hostent	   *pHostEntry;
char			   *pDestIp;
void			   *pUser;
struct sockaddr_in	sDest;
unsigned int		uAddr;
int					iDone;
int					iPos;
int					iOk;



	if(iCount<=0)												// Zuwenig Eintrge angegeben
		{
		return FALSE;
		}


	for(iPos=0;iPos<iCount;iPos++)								// Socket-Handels lschen
		{
		pUser = pData[iPos].pUserData;
		memset( pData+iPos,0,sizeof(PING_DATA));
		pData[iPos].uSocket   = INVALID_SOCKET;
		pData[iPos].pUserData = pUser;
		}


	if(!iInitDone)												// IP-Stack initialisieren
		{
		if(WSAStartup(MAKEWORD(2,1),&sWsaData) != 0)
			{
			pData->uSocket=INVALID_SOCKET;
			return FALSE;
			}

		iInitDone = 1;
		}


	   pData[0].uSocket = socket(AF_INET,SOCK_RAW,IPPROTO_ICMP);// Sende-Socket erzeugen
	if(pData[0].uSocket == INVALID_SOCKET)
		{
		return FALSE;
		}

	iTimeout = 1000;

	   iOk  = setsockopt(pData[0].uSocket,SOL_SOCKET,SO_RCVTIMEO,(char*)&iTimeout,sizeof(iTimeout));
	if(iOk == SOCKET_ERROR)
		{
		closesocket(pData[0].uSocket);
		pData[iPos].uSocket=INVALID_SOCKET;
		return FALSE;
		}

	iTimeout = 1000;

	   iOk =  setsockopt(pData[0].uSocket,SOL_SOCKET,SO_SNDTIMEO,(char*)&iTimeout,sizeof(iTimeout));
	if(iOk == SOCKET_ERROR)
		{
		closesocket(pData[0].uSocket);
		pData[0].uSocket=INVALID_SOCKET;
		return FALSE;
		}


	iDone = 0;


	for(iPos=0;iPos<iCount;iPos++)								// Alle Eintrge abarbeiten
		{
		if(*pHostName==0)break;

		pData[iPos].uNextValid = 1;
		memset(&sDest,0,sizeof(sDest));							// Zieladressen erzeugen

		pHostEntry = gethostbyname(pHostName);
		uAddr	   = 0;

		if(!pHostEntry)
			{
			uAddr = inet_addr(pHostName);
			}

		if(!pHostEntry &&  uAddr==INADDR_NONE)
			{
			continue;
			}

		if(pHostEntry!=NULL)
			{
			memcpy(&(sDest.sin_addr),pHostEntry->h_addr,pHostEntry->h_length);
			}
		else{
			sDest.sin_addr.s_addr = uAddr;
			}

		if(pHostEntry)
			{
			sDest.sin_family = pHostEntry->h_addrtype;
			}
		else{
			sDest.sin_family = AF_INET;
			}

		   pDestIp = inet_ntoa(sDest.sin_addr);
		if(pDestIp)
			{
			strncpy(pData[iPos].cDestIp,pDestIp,sizeof(pData[iPos].cDestIp));
			}
		else{
			pData[iPos].cDestIp[0]=0;
			}

		pData[iPos].uSocket		 = pData[0].uSocket;
		pData[iPos].iDataSize    = DEF_PACKET_SIZE + sizeof(IcmpHeader);
		pData[iPos].pSendData    = new char[MAX_PACKET];
		pData[iPos].pRecvData    = new char[MAX_PACKET];
		pData[iPos].pDestination = new char[sizeof(sDest)];
		pData[iPos].iDestLen	 = 10;

		memcpy(pData[iPos].pDestination,&sDest,sizeof(sDest));
		memset(pData[iPos].pSendData,0,MAX_PACKET);
		memset(pData[iPos].pRecvData,0,MAX_PACKET);


		PingFilIcmpData(pData[iPos].pSendData,pData[iPos].iDataSize);

		pHostName += strlen(pHostName);
		pHostName += 1;
		iDone     += 1;
		}



	if(iDone<=0)												// Konnte mindestens eine Addresse erzeugt werden
		{
		closesocket(pData[iPos].uSocket);
		pData[0].uSocket    = INVALID_SOCKET;
		pData[0].uNextValid = 0;
		return FALSE;
		}

	pData[iPos-1].uNextValid = 0;



return TRUE;
}


//*****************************************************************************
//*
//*		PingExit
//*
//*****************************************************************************
//	Beendet die Socket-Verbindung und deinitialisiert die Daten
//	pData	: Zeiger auf die Ping Daten
//	Ergibt 1 wenn ok sonst 0
BOOL PingExit(PING_DATA *pData)
{
int		iRet=TRUE;
int		iPos;




	for(iPos=0;;iPos++)
		{
		if(pData->uSocket==INVALID_SOCKET)							// Ist das Socket offen
			{
			if(iPos==0)iRet = FALSE;
			}
		else{
			closesocket(pData->uSocket);
			pData->uSocket=INVALID_SOCKET;
			}

		if(pData->pRecvData)
			{
			delete pData->pRecvData;
			pData->pRecvData = 0;
			}

		if(pData->pSendData)
			{
			delete pData->pSendData;
			pData->pSendData = 0;
			}

		if(pData->pDestination)
			{
			delete pData->pDestination;
			pData->pDestination = 0;
			}

		if(!pData->uNextValid)break;								// Gibt es weitere Eintrge

		pData++;
		}



return iRet;
}

//*****************************************************************************
//*
//*		PingSend
//*
//*****************************************************************************
//	Sendet einen Ping ab
//	pData		: Zeiger auf die Ping Daten
//	uTimeout	: Ist die Wartezeit auf die Antwort in ms
//				  Bei 0 wird auf die Antwort nicht gewartet
//	iCount		: Ist die Anzahl der zu sendenten Blcke.
//	Ergibt  >=0 Ping Zeit in ms
//			 -1  Sendefehler
//			 -2  Empfangsfehler
//			 -3	 Falscher Telegrammtype empfangen
//			 -4	 Packet von anderm Process
//			 -5  Auf die Antwort wurde nicht gewartet
//	Gespeicherte Werte:
//		pData[].uRecvTime	= Ping-Zeit in ms vom letzten empfangenen Paket.
//		pData[].iResponse   = Ping-Zeit oder Fehlercode (-1..-5) vom letzten Aufruf
//		pData[].iRecvFlag   = Wurde beim letzten Aufruf etwas empfangen
//		pData[].iRecvCount  = Anzahl der empfangenen Pakete
int PingSend(PING_DATA *pData,unsigned uTimeout,int iCount)
{
unsigned			uSocket;
IcmpHeader		   *pHeader;
int					iWrote;
int					iDone;
int					iMax;
int					iPos;



	uSocket	 = pData->uSocket;

	if(uSocket==INVALID_SOCKET)										// Ist das Socket in Ordnung
		{
		for(iPos=0;iCount>0;iPos++,iCount--)
			{
				pData[iPos].iResponse=-1;
				pData[iPos].iRecvFlag= 0;
			if(!pData[iPos].uNextValid)break;
			}

		return -1;
		}


	iDone = 0;


	for(iPos=0;iCount>0;iPos++,iCount--)							// Alle Pakete senden
		{
		if(!pData[iPos].pDestination)
			{
			if(!pData[iPos].uNextValid)break;
			continue;
			}

		pHeader				 = (IcmpHeader*)(pData[iPos].pSendData);
		pHeader->usChkSum    = 0;
		pHeader->uiTimeStamp = GetTickCount();
		pHeader->usSequ      = pData[iPos].uNumber++;
		pHeader->usChkSum    = PingCheckSum((USHORT*)pData[iPos].pSendData, pData[iPos].iDataSize);

		   iWrote = sendto(uSocket,pData[iPos].pSendData,pData[iPos].iDataSize,0,(struct sockaddr*)pData[iPos].pDestination,sizeof(struct sockaddr_in));
		if(iWrote == SOCKET_ERROR)
			{
			pData[iPos].iResponse = -1;
			pData[iPos].uNumber  -=  1;
			}
		else{
			pData[iPos].iSendFlag =  1;
			iDone++;
			}


		if(!pData[iPos].uNextValid)break;
		}


	if(!iDone)return -1;											// Konnte nichts versendet werden


	iMax = iPos+1;


	if(!uTimeout)													// Soll auf die Antwort gewartet werden ?
		{
		for(iPos=0;iPos<iMax;iPos++)
			{
			pData[iPos].iRecvFlag= 0;
			pData[iPos].iResponse=-5;
			}

		return -5;
		}


return PingResv(pData,uTimeout);
}

//*****************************************************************************
//*
//*		PingResv
//*
//*****************************************************************************
//	Wartet auf einen Ping-Empfang
//	pData		: Zeiger auf die Ping Daten
//	uTimeout	: Ist die Wartezeit auf die Antwort in ms
//				  Bei 0 wird auf die Antwort nicht gewartet
//	Ergibt  >=0 Ping-Zeit in ms
//			 -1  Sendefehler
//			 -2  Empfangsfehler
//			 -3	 Falscher Telegrammtype empfangen
//			 -4	 Packet von anderm Process
//			 -5  Auf die Antwort wurde nicht gewartet
//	Gespeicherte Werte:
//		pData[].uRecvTime	= Ping-Zeit in ms vom letzten empfangenen Paket.
//		pData[].iResponse   = Ping-Zeit oder Fehlercode (-1..-5) vom letzten Aufruf
//		pData[].iRecvFlag   = Wurde beim letzten Aufruf etwas empfangen
//		pData[].iRecvCount  = Anzahl der empfangenen Pakete
int PingResv(PING_DATA *pData,unsigned uTimeout)
{
char				cData[MAX_PACKET];
unsigned			uSocket;
unsigned			uResvTime;
unsigned			uStartTime;
struct timeval		sTimeout;
struct sockaddr_in	sFrom;
fd_set				sRead;
char			   *pFromIp;
int					iFromLen;
int					iOther;
int					iRead;
int					iRecv;
int					iDiff;
int					iFlag;
int					iFail;
int					iNum;
int					iMax;
int					iVal;
int					iRet;
int					iPos;



	uSocket	 = pData->uSocket;
	iFromLen = sizeof(sFrom);

	if(uSocket==INVALID_SOCKET)									// Ist das Socket in Ordnung
		{
		for(iPos=0;;iPos++)
			{
				pData[iPos].iResponse=-1;
				pData[iPos].iRecvFlag= 0;
			if(!pData[iPos].uNextValid)break;
			}

		return -1;
		}


	for(iPos=0;;iPos++)											// Alle auf nicht empfangen
		{
		    pData[iPos].iResponse=-2;
			pData[iPos].iRecvFlag= 0;
		if(!pData[iPos].uNextValid)break;
		}

	iMax = iPos+1;


	iRet		= -2;
	iRecv		=  0;
	iFail		=  0;
	iOther		=  0;
	uStartTime	=  GetTickCount();


	for(;;)
		{
		if(uTimeout)												// Auf einen Empfang warten
			{
			   iDiff = uStartTime+uTimeout-GetTickCount();
			if(iDiff<=0){iDiff=1;uTimeout=0;}
			sTimeout.tv_sec  = (iDiff/1000);
			sTimeout.tv_usec = (iDiff%1000)*1000;
			}
		else{														// Nicht warten
			sTimeout.tv_sec  = 0;
			sTimeout.tv_usec = 0;
			}

		FD_ZERO(&sRead);
		FD_SET(pData->uSocket,&sRead);

		   iNum=select(pData->uSocket+1,&sRead,0,0,&sTimeout);
		if(iNum<=0)													// Wurde nichts empfangen
			{
			break;
			}


		uResvTime = GetTickCount();


		   iRead =recvfrom(pData->uSocket,cData,sizeof(cData),0,(struct sockaddr*)&sFrom,&iFromLen);
		if(iRead==SOCKET_ERROR)
			{
			if(iRecv==0)return -2;
			if(uTimeout==0)break;
			uTimeout=0;
			continue;
			}


		for(iPos=0;iPos<iMax;iPos++)								// Passt das Paket zu einem Block
			{
			if(pData[iPos].iSendFlag)
			if(iFromLen>=pData[iPos].iDestLen)
			if(PingCmpIpAddr(&sFrom,(struct sockaddr_in*)pData[iPos].pDestination)==0)
				{
				iVal = PingDecodeResponse(cData,iRead,&sFrom,&pData[iPos].uRecvNumber);

				     if(iVal==-3)iFail=1;
				else if(iVal==-4)iOther++;
				else{
					if(!pData[iPos].iRecvFlag)iRecv++;

					memcpy(pData[iPos].pRecvData,cData,iRead);

					pData[iPos].uRecvTime	= uResvTime-iVal;
					pData[iPos].iResponse   = uResvTime-iVal;
					pData[iPos].iRecvFlag   = 1;
					pData[iPos].iRecvCount += 1;

					if(iRet==-2)
						{
						iRet  = pData[iPos].iResponse;
						}

					   pFromIp = inet_ntoa(sFrom.sin_addr);
					if(pFromIp)
						{
						strncpy(pData[iPos].cFromIp,pFromIp,sizeof(pData[iPos].cFromIp));
						}

					iRecv++;
					iFlag=1;
					}
				}
			}

		if(!iFlag)
			{
			iOther++;
			}

		if(iRecv>=iMax)												// Wurden alle Pakete empfangen
			{
			uTimeout=0;
			}
		}


	if(iRet==-2)
		{
		if(iFail   )iRet = -3;
		if(iOther>0)iRet = -4;
		}


return iRet;
}
